//June 28, 2001. Made it legal for this buffer to be
//created with zero capacity. Before, perhaps more
//rightfully, an IllegalArgumentException would be
//thrown if the buffer was given a capacity equal to zero.
//This was done to cheaply fix a serious bug in the
//SearchDocument class. The SearchDocument class was
//depending on it being OK to have a zero length buffer.
//It wasn't, before the change.
//
//August 11, 2001. Fixed a very serious bug introduced in the last change.
//The incorrect field was used in several conditionals within the 
//CircularStringBuffer class, leading instances to always return the
//empty string from toString().
//This problem has been fixed. This problem was pointed out by

/**
 * Limited length buffer.
 * <br>
 * Once <code>bufferSize</code> char have been appended to the
 * buffer, char are discarded from the front of the buffer to make
 * room at the rear.
 *
 * @author David Faden <a href="mailto:gilbertnews@hotmail.com">gilbertnews@hotmail.com</a>
 * @version CircularStringBuffer 1.2 August 11, 2001
 */
public class CircularStringBuffer {
    
    /**
     * Where the next char should be inserted.
     */
    private int insertionIndex;
    
    /**
     * The number of char in the buffer.
     */
    private int size;
    
    /**
     * The maximum number of char that this buffer will accept.
     */
    private int bufferSize;
    
    /**
     * Storage for the buffer.
     */
    private char[] contents;

    /**
     * Create a buffer that will hold up to <code>bufferSize</code>
     * char.
     * <br>
     * Once <code>bufferSize</code> char have been appended to the
     * buffer, char are discarded from the front of the buffer to make
     * room at the rear.
     * <br>
     * If <code>bufferSize</code> is set to be less than zero, this
     * buffer becomes a bottomless pit, gobbling up anything appended
     * to it, but in the end still only returning an empty String.
     *
     * @param bufferSize The maximum size of the buffer.
     * @throws IllegalArgumentException If <code>bufferSize</code> &lt; 0.
     */
    public CircularStringBuffer(int bufferSize) {
        if (bufferSize < 0)
            throw new IllegalArgumentException("bufferSize < 0");
    
        this.bufferSize = bufferSize;
        contents = new char[bufferSize];
        size = 0;
        insertionIndex = 0;
    }
    
    /**
     * Append a char to the end of the buffer.
     * <br>
     * Once <code>bufferSize</code> char have been appended to the
     * buffer, char are discarded from the front of the buffer to make
     * room at the rear.
     *
     * @param c Char to append.
     */
    public void append(char c) {
        if (bufferSize != 0) {
            contents[insertionIndex] = c;
            if (size < bufferSize)
                size++;
            insertionIndex++;
            insertionIndex %= bufferSize;
        }
    }
    
    /**
     * The number of char in the buffer.
     *
     * @return The number of char in the buffer.
     */
    public int getSize() {
        return size;
    }
    
    /**
     * The maximum number of char that this buffer will accept.
     */
    public int getBufferSize() {
        return bufferSize;
    }
    
    /**
     * Convert the contents of this buffer to a String.
     *
     * @return The String equivalent of this buffer.
     */
    public String toString() {
        if (bufferSize != 0) {
            if (size < bufferSize || insertionIndex == 0) {
                return new String(contents, 0, size);
            }
            else {
                StringBuffer buffer = new StringBuffer(bufferSize);
                buffer.append(contents, insertionIndex, bufferSize - insertionIndex);
                buffer.append(contents, 0, insertionIndex);
                return buffer.toString();
            }    
         }
         else {
            return "";
         }
    } 
}
