import java.awt.*;

// March 13, 2001: Added comments. Changed changeBy to 
// work properly with negative arguments, which its name suggested
// it should do. Added code to ensure that all of the arguments
// to the constructor are valid.
// Version 1.2
//
// March 24, 2001: Added more comments. Added code to setProgressBarVisible(boolean)
// method to call repaint().
// Version 1.3

/**
 * A class to graphically display the percent of a task completed.
 *
 * @author David Faden <a href="mailto:gilbertnews@hotmail.com">gilbertnews@hotmail.com</a>
 * @version ProgressBar 1.3 March 24, 2001
 */
public class ProgressBar extends Canvas {

    /**
     * True if the ProgressBar should be made visible.
     */
    private boolean progressBarVisible;
    
    /**
     * True if text message should be drawn over the ProgressBar.
     */
    private boolean percentMessageVisible;

    /**
     * The color of the portion of the ProgressBar representing what has been completed.
     */
    private Color completedPartColor;
    
    /**
     * The color of the portion of the ProgressBar representing what remains to be completed.
     */
    private Color remainingPartColor;
    
    /**
     * The color of the text displayed in the ProgressBar.
     */
    private Color textColor;
    
    /**
     * Message displayed with the percent done in the ProgressBar.
     */
    private String message;
    
    /**
     * The maximum value -- the value that currentValue will have when the task is complete.
     */
    private int maximumValue;
    
    /**
     * The current value representing the completed portion of the task.
     */
    private int currentValue;
 
    /**
     * Create a new ProgressBar.
     * <br>
     * To explicitly set the ProgressBar's dimensions, call
     * Component.resize(int, int) or Component.setSize(Dimension)
     * on the ProgressBar, the choice of method
     * depending on what version of Java you're targeting (1.0 or
     * 1.1 and above, respectively).
     * <br>
     * When the current value of the ProgressBar hits the value
     * given in maximumValue, it displays 100%.
     * <br>
     * The background of the ProgressBar, the area representing
     * the portion of the task yet to be completed is colored
     * remainingPartColor.
     * <br>
     * The part of the task completed is represented by a bar
     * extending from the left side towards the right, colored
     * completedPartColor. Most applications, it seems, use
     * a darker shade to represent the completed part and
     * a lighter shade the part still to be done.
     * <br>
     * By default, as the value of the ProgressBar changes,
     * text is drawn in the ProgressBar showing what
     * percent the current value is of the maximum value.
     * This text is preceded by the String in <code>message</code>
     * and is drawn <code>textColor</code> color.
     * <br>
     * All of the parameters must be non-null.
     *
     * @param maximumValue Value greater than zero for completion.
     * @param completedPartColor The color of the portion of the 
     * ProgressBar representing what has been completed.
     * @param remainingPartColor The color of the portion of the ProgressBar
     * representing what remains to be completed.
     * @param textColor The color of the text displayed in the ProgressBar.
     * @param message Message displayed with the percent done in the ProgressBar.
     */
    public ProgressBar(int maximumValue,
                       Color completedPartColor, 
                       Color remainingPartColor,
                       Color textColor,
                       String message) {
            if (completedPartColor == null)
                throw new IllegalArgumentException("null completedPartColor");
            this.completedPartColor = completedPartColor;
            
            if (remainingPartColor == null)
                throw new IllegalArgumentException("null remainingPartColor");
            this.remainingPartColor = remainingPartColor;
            
            if (textColor == null)
                throw new IllegalArgumentException("null textColor");
            this.textColor = textColor;
            
            if (maximumValue <= 0)
                throw new IllegalArgumentException("maximumValue < 0");
            this.maximumValue = maximumValue;
            
            if (message == null)
                throw new IllegalArgumentException("null message");
            this.message = message;
            
            progressBarVisible = true;
            percentMessageVisible = true;
    }
    
    public Color getCompletedPartColor() {
        return completedPartColor;
    }
     
    public void setCompletedPartColor(Color completedPartColor) {
        this.completedPartColor = completedPartColor;
        repaint();
    }
    
    public Color getRemainingPartColor() {
        return remainingPartColor;
    }
    
    public void setRemainingPartColor(Color remainingPartColor) {
        this.remainingPartColor = remainingPartColor;
        repaint();
    }
  
    public Color getTextColor() {
        return textColor;
    }
  
    public void setTextColor(Color textColor) {
        this.textColor = textColor;
        repaint();
    }

    public boolean isPercentMessageVisible() {
        return percentMessageVisible;
    }
    
    public void setPercentMessageVisible(boolean percentMessageVisible) {
        this.percentMessageVisible = percentMessageVisible;
        repaint();
    }
    
    /**
     * Get the message displayed with the percent done in the ProgressBar.
     * 
     * @return String holding the message displayed with the percent done in 
     * the ProgressBar.
     */  
    public String getMessage() {
        return message;
    }

    /**
     * Set the message displayed with the percent done in the ProgressBar.
     * <br>
     * The ProgressBar is repainted with the new message.
     * 
     * @param message String holding the message displayed with the percent done in 
     * the ProgressBar.
     */     
    public void setMessage(String message) {
        this.message = message;
        repaint();
    }
  
    public void setMaximumValue(int maximumValue) {
        this.maximumValue = maximumValue;
        repaint();
    }
    
    public int getMaximumValue() {
        return maximumValue;
    }
    
    /**
     * Return the current value representing the completed portion of the task.
     *
     * @return The current value representing the completed portion of the task.
     */
    public int getCurrentValue() {
        return currentValue;
    }
    
    /**
     * Set the current value representing the completed portion of the task.
     * <br>
     * The ProgressBar is repainted to reflect the new current value.
     *
     * @param currentValue Value representing the completed portion.
     */
    public void setCurrentValue(int currentValue) {
        this.currentValue = currentValue;
        repaint();
    }
    
    public double getPercentDone() {
        return (100.0D * currentValue) / maximumValue;
    }
    
    /**
     * Change the current value of this ProgressBar
     * by the given amount.
     * <br>
     * If the new value is greater than the maximum,
     * the current value is set to the maximum.
     * <br>
     * If the new value is less than zero, the current value
     * is set to zero.
     *
     * @param amount The change in the current value.
     * @return The new current value.
     */
    public int changeBy(int amount) {
        int newValue = currentValue + amount;
    
        if (newValue > maximumValue) 
            newValue = maximumValue;
        else if (newValue < 0)
            newValue = 0;
            
        currentValue = newValue;
            
        repaint();
        return currentValue;
    }
  
    /**
     * Set the current value of this ProgressBar to zero.
     * That is, set this ProgressBar to show that zero percent
     * of the task has been completed.
     * <br>
     * A call to this method causes the ProgressBar to be repainted.
     */
    public void reset() {
        currentValue = 0;
        repaint();
    }
    
    public void paint(Graphics g) {
        Dimension d = size(); //Deprecated
        int width = d.width;
        int height = d.height;
    
        if (progressBarVisible) {
            g.setColor(remainingPartColor);
            g.fillRect(0, 0, width, height);
            if (currentValue > 0) {
                float fractionDone = (float) currentValue / (float) maximumValue;
                if  (fractionDone > 1.0f) 
                    fractionDone = 1.0f;
                int drawWidth = (int) ((float) (width - 2) * fractionDone);
                g.setColor(completedPartColor);
                g.fillRect(1, 1, drawWidth, height - 2);
                int percentDone = (int) (100.0 * fractionDone);
                if (percentMessageVisible) {
                    g.setColor(textColor);
                    g.drawString(message + " " + percentDone + "%", 4, height - 5);
                    //XXX! Internationalization issue.
                }
            }
        }
        else {
            g.setColor(getParent().getBackground());
            g.fillRect(0, 0, width, height);
        }
    }

    /**
     * Control whether the ProgressBar should be drawn.
     * <br>
     * Pass true to this method to have the ProgressBar drawn,
     * false to hide it.
     * <br>
     * If the ProgressBar is made invisible, the ProgressBar
     * is overdrawn with the background color of the ProgressBar's
     * parent Component.
     * <br>
     * A call to this method causes the ProgressBar to be repainted.
     *
     * @param visible True if the ProgressBar should be made visible.
     */  
    public void setProgressBarVisible(boolean visible) {
        progressBarVisible = visible;
        repaint();
    }

    /**
     * Returns true if this ProgressBar is visible, false, otherwise.
     *
     * @return True if this ProgressBar is visible, false, otherwise.
     */    
    public boolean isProgressBarVisible() {
        return progressBarVisible;
    }
}
