/*
 * SearchSieve.java. Copyright (c) 1999 - 2001, David Faden. All rights reserved.
 *
 * This code is distributed without warranty of any kind.
 */

//A partial modification record:
//
// August 16, 2000: Corrected some inaccurate comments.
//
// August 17, 2000: Removed method isSpace(char). Added the 
// isWordChar(int) method to be used as negative replacement.
// isWordChar uses table look-up rather than a loop so it should be
// substantially faster. Added method setWordChar(char, boolean) 
// to match isWordChar(int).
// I consider this to be an interim version of isWordChar - the final
// one will use some sort of hash and correctly handle the full range of
// Unicode characters.
//
// August 17, 2000: changed the names of the methods numberOfMatches() and patternLength()
// to getNumOfMatches() and getKeyLength() respectively. Documented the undocumented.
//
// December 28, 2000: added some more documentation, shifted some conditionals around
// to better reflect the expected occurrence of conditions.
//
// March 11, 2001: Changed many of the variable names to be more intuitive.
// Added code to keep track of the "position" or "index" of the SearchSieve
// in relation to the stream of char fed to it. Changed the constructor to take
// a String as its argument rather than a char array. Strings are safer.
// Strings are more convenient to handle for the caller. Using a String better
// fits the Java way of thinking.
//
// March 13, 2001: Stripped out code that allowed any two white space
// char to match -- a tab would be considered to match a carriage return.
// I guestimated that the cost of this feature outweighed the benefits. 
// Also, this behavior will not be so easy (or possible) to implement if 
// and when SearchSieves are switched to use the Knuth-Morris-Pratt 
// algorithm internally. See addChar().
//

/**
 * The SearchSieve class sifts for a pattern in a sequence
 * of Java (Unicode) characters fed to the sieve with addChar.
 *
 * @author David Faden
 * @version SearchSieve 2.1 March 13, 2001
 */
public final class SearchSieve {
    
    /**
     * Current position in the pattern.
     */
    private int patternIndex;
    
    /**
     * Current index in the stream of characters fed to this SearchSieve.
     */
    private int textIndex;
    
    /**
     * Index of the beginning of the most recent match.
     */ 
    private int mostRecentMatchIndex;
    
    /**
     * String representation of the search pattern.
     */
    private String patternString;
    
    /**
     * The pattern to search for.
     */
    private char[] pattern;
    
    /**
     * The number of characters in the pattern.
     */
    private int patternLength;
    
    /**
     * Number of matches found by this SearchSieve -- reset to zero by reset() or setPattern().
     */
    private int numberOfMatches;
    
    /**
     * Determines whether a match can be part of a larger word.
     */
    private boolean exactMatchesOnly;
  
    /**
     * The previous character to be examined.
     */
    private char prevc;
  
    /**
     * Store characters for backtracking.
     */
    private char[] buffer;
    
    /**
     * The number of characters in the buffer.
     */
    private int bufferLength;
    
    /**
     * Determines whether this SearchSieve is buffering to backtrack.
     */
    private boolean buffering;
    
    /**
     * Storage for swaps between char arrays.
     */
    private char[] swapSpace;
  
    /**
     * Word character table used for fast look-up.
     */
    private static boolean[] wordCharTable;
    //XXX! This table excludes multi-byte characters. A search for a better
    //solution is under way.
    static {
        wordCharTable = new boolean[256];
        for (int i = 0; i < 256; i++) {
            if (i >= 'A' && i <= 'Z' ||
                i >= 'a' && i <= 'z' ||
                i >= '0' && i <= '9' ||
                i >= 0xC0) //LATIN CAPITAL LETTER A WITH GRAVE and up
                wordCharTable[i] = true;
            else
                wordCharTable[i] = false;
                //I believe that the Java Lang Spec calls for
                //booleans to have an uninitialized value of false anyway but...
        }
    }
  
    /**
     * Create a new SearchSieve.
     *
     * @param patternString - The pattern to search for.
     * @param exactMatchesOnly - If true, matches must be surrounded by non-word characters.
     */
    public SearchSieve(String patternString, boolean exactMatchesOnly) {
        setPattern(patternString, exactMatchesOnly);
    }
  
    /**
     * Set the pattern this SearchSieve should search for.
     *
     * @param pattern String containing the pattern to search for.
     * @param exactMatchesOnly True if only matches surrounded by non-word
     *    characters should be accepted.
     */
    public final void setPattern(String patternString, boolean exactMatchesOnly) {
        this.exactMatchesOnly = exactMatchesOnly;
        this.patternString = patternString;
        pattern = patternString.toCharArray();
        patternLength = pattern.length;
        buffer = new char[patternLength];
        swapSpace = new char[patternLength]; 
        //swapSpace is created here to avoid the overhead of recreating it 
        //each time it is needed.
    
        reset();
    }
    
    /**
     * Reset this SearchSieve so that it acts as if addChar() has not
     * been called since the pattern was last set. The count of matches
     * is set back to zero.
     */
    public final void reset() {
        textIndex = -1;
        mostRecentMatchIndex = -1;
        patternIndex = 0;
        numberOfMatches = 0;
        prevc = '\0'; //null according to the Unicode table, nul according to others
        bufferLength = 0; //clear the buffer
        buffering = false;
    }
      
    /**
     * Return the String representation of the pattern
     * being searched for, or the pattern for which this
     * SearchSieve is searching, if you will.
     *
     * @return The String representation of the search pattern.
     */
    public final String getPattern() {
        return patternString;
    }
    
    /**
     * Return the number of characters in pattern (pattern) being
     * searched for.
     *
     * @return The number of characters in pattern (pattern) being
     * searched for.
     */
    public final int getPatternLength() {
        //XXX! Is useful to have this method? I think it will be
        //in case SearchSieves are later used as part of a wildcard
        //matching class.
        return patternLength;
    }
  
    /**
     * Current index in the stream of characters fed to this SearchSieve.
     * 
     * @return Current index in the stream of characters fed to this SearchSieve.
     */
    public final int getIndexInText() {
        return textIndex;
    }
    
    /**
     * Index of the beginning of the most recent match in 
     * the stream of characters fed to this SearchSieve.
     * <br>
     * Even if this SearchSieve has been set to search for
     * only exact matches, this method still returns the index
     * of the first char of the pattern in the text stream, not
     * the non-word char preceding it.
     *
     * @return Index of the most recent match.
     */
    public final int getMostRecentMatchIndex() {
        return mostRecentMatchIndex;
    }
  
    /**
     * Returns the number of occurrences of the current pattern 
     * this SearchSieve has seen.
     * 
     * @return The number of occurrences of the current pattern 
     * this SearchSieve has seen.
     */
    public final int getNumberOfMatches() {
        return numberOfMatches;
    }
    
    /**
     * Returns true if the sequence of chars added to the SearchSieve,
     * including <code>c</code>, constitutes a match. For exactMatchesOnly searches,
     * the last char of a match must be a non-word char.
     * <br>
     * If the pattern is of zero length, this method always returns false.
     *
     * @param c The next char in the text stream.
     * @return True if <code>c</code> completes a match -- false, otherwise.
     */
    public final boolean addChar(char c) {
        boolean matchFound = false; //Left unchanged if nothing's found.
        boolean considerEatingBuffer = false;
        
        textIndex++;
        
        if (patternLength < 1)
            return false; //XXX! Is this the best behavior?
        
        if (exactMatchesOnly) {
            if (patternIndex == 0) {
                if (c == pattern[0] && !isWordChar(prevc)) 
                    patternIndex++;
            }
            else if (patternIndex != patternLength) {
                if (c != pattern[patternIndex]) { //found no match
                    patternIndex = 0;
                    considerEatingBuffer = true;
                }
                else {
                    patternIndex++;
                }
            }
            else { //all the characters of the pattern matched
                if (!isWordChar(c)) { //found an "exactMatchesOnly" match
                    matchFound = true;
                }
                patternIndex = 0;
                considerEatingBuffer = true;
            }
        }
        //Inexact match check
        else {
            if (c != pattern[patternIndex]) { 
                    //No match
                    patternIndex = 0;
                    considerEatingBuffer = true;
            }
            else //Input char matches pattern
                patternIndex++;
            
            if (patternIndex == patternLength) { //All the characters matched!
                patternIndex = 0;
                considerEatingBuffer = true;
                matchFound = true;
            }
        }
        
        //Note: patternIndex of 1 indicates that one char match has been found
        if (patternIndex > 1 && !buffering && c == pattern[0]) {
            if (!exactMatchesOnly || exactMatchesOnly && !isWordChar(prevc)) {
                buffering = true;
                buffer[0] = c;
                bufferLength = 1;
            }
        }
        else if (buffering) {
            if (c != pattern[bufferLength]) {
                bufferLength = 0;
                buffering = false;
            }
            else {
                buffer[bufferLength] = c;
                bufferLength++;
            }
        }
        
        if (considerEatingBuffer && bufferLength > 0) 
            eatTheBuffer(); //Feed the buffer back into the
            //search system on failures and finds
            
        prevc = c;
        
        if (matchFound) {
            numberOfMatches++;
            mostRecentMatchIndex = textIndex - patternLength + 1;
            if (exactMatchesOnly)
                mostRecentMatchIndex += -1;
        }
        
        return matchFound;
    }
  
  	public String toString() {
  		return getPattern();
  	}
  
    /**
     * Feed the buffer through addChar(char) as if it were being externally entered.
     */
    private final void eatTheBuffer() {
        //For historical interest, I have preserved the original comments below:
        //Feed the buffer through addChar(char) as if it were being externally entered.
        //No match can ever be found in the buffer alone, so there is
        //no need to check the return of addChar(char).
        //Huge problem with this method which will cause fatal error...
        //Buffer itself may contain a second ambiguous char...
        //However, if the buffer doesn't match up, it is overwritten
        //When the next call to eatTheBuffer is made
        //Idea, what if swapSpace is made internal to eatTheBuffer?
        //This could lead to a major memory leak, though...
        //Fixed...breath easy now!
        //XXX! If this step were replaced with a look up table,
        //SearchSieves would roughly implement the Knuth-Morris-Pratt
        //algorithm...
        
        //Reset the search
        patternIndex = 0;
        
        System.arraycopy(buffer, 0, swapSpace, 0, bufferLength);
        int tempLength = bufferLength; //Save the value of bufferLength
        
        //Reset the buffer
        bufferLength = 0; 
        buffering = false;
    
        //Don't bother checking addChar(char)'s return value -- 
        //it can never be true in this case.
        for (int i = 0; i < tempLength; i++) 
            addChar(swapSpace[i]);
    }

  
    /**
     * Returns true if SearchSieves consider <code>c</code> to
     * be part of a word. 
     * <p>
     * Currently, this method is limited to 
     * characters &lt; 256 and will return false for all
     * values out of range. This failing will only affect
     * "exactMatchesOnly" searches, potentially returning false positives.
     *
     * @param c  The char to test.
     * @return True if <code>c</code> is a word char.
     */
    public static final boolean isWordChar(int c) {
        if (c < 0 || c > 255)
            return false;
        return wordCharTable[c];
    }
    
    /**
     * Set whether SearchSieves should consider <code>c</code>
     * to be part of a word.
     * <p>
     * Currently, this method is limited to char &lt; 256 and will
     * will throw an IllegalArgumentException for out of range char.
     *
     * @param c  The char to set.
     * @param isWord  A true value will cause SearchSieves to consider c part of a word.
     * @exception IllegalArgumentException  Thrown if c > 255.
     */
    public static final void setWordChar(char c, boolean isWord) {
        int i = c;
        //XXX! Remember to chop this section once the code is fully
        //internationalized.
        if (i == 0 || i > 255)
            throw new IllegalArgumentException("Out of range.");
        wordCharTable[i] = isWord;
    }
  
}
